# python standard

# third-party imports
from flask import abort, Flask, render_template
from flask_bootstrap import Bootstrap
from flask_login import LoginManager
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy

# local imports
from webapp.config import app_config

db = SQLAlchemy()
login_manager = LoginManager()


def create_app(config_name):
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_object(app_config[config_name])
    app.config.from_pyfile('config.py')

    Bootstrap(app)
    db.init_app(app)
    login_manager.init_app(app)
    login_manager.login_message = "You must be logged in to access this page."
    login_manager.login_view = "auth.login"
    migrate = Migrate(app, db)

    from webapp.app import models

    from .admin import admin as admin_blueprint
    app.register_blueprint(admin_blueprint, url_prefix='/admin')

    from .auth import auth as auth_blueprint
    app.register_blueprint(auth_blueprint)

    from .home import home as home_blueprint
    app.register_blueprint(home_blueprint)

    @app.errorhandler(403)
    def forbidden(error):
        return render_template('errors/403.html', title='Forbidden'), 403

    @app.errorhandler(404)
    def page_not_found(error):
        return render_template('errors/404.html', title='Page Not Found'), 404

    @app.errorhandler(500)
    def internal_server_error(error):
        return render_template('errors/500.html', title='Server Error'), 500

    @app.route('/500')
    def error():
        abort(500)

    return app



# -*- coding: utf-8 -*-
"""
Test workers, how they should extract stuff, how they should parse .csv
files, and how they should store it in the database.
"""

# python standard
import unittest
import logging
from unittest.mock import patch

# third-party imports
import pandas as pd
import numpy as np

# local imports
from utils.bucket_helper import BucketHelper
from workers.worker import Worker, DcmWorker, DspWorker

logging.disable(logging.CRITICAL)


class TestWorkers(unittest.TestCase):
    def setUp(self):
        """
        Some examples of information that could be found in the bucket
        """

        # a fake list of files
        self.fake_list = [
            {
                'name': 'archive/',
                'contentType': 'application/x-www-form-urlencoded',
                'size': '0'
            }, {
                'name': 'dbm.csv',
                'contentType': 'text/csv',
                'size': '21645'
            }, {
                'name': 'dcm.csv',
                'contentType': 'text/csv',
                'size': '48297'
            }, {
                'name': 'mediamath.csv',
                'contentType': 'text/csv',
                'size': '22989'
            }
        ]

        # first two lines of a good dcm's file
        self.good_dcm = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": "48988.5",
            "clicks": "118",
            "reach": "48.24"
        }, {
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "198706",
            "placement": "dbm_youtube",
            "impressions": "55509.0",
            "clicks": "116",
            "reach": "42.49"
        }])

        # first two line of a file with data missing, will be filled with zero
        self.missing_fine_dcm = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": np.nan,
            "clicks": "",
            "reach": ""
        }, {
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "198706",
            "placement": "dbm_youtube",
            "impressions": "",
            "clicks": "116",
            "reach": "42.49"
        }])

        # first two line of a file with data missing, in this case, dimensions
        # are missing
        self.missing_bad_dcm = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": "48988.5",
            "clicks": "118",
            "reach": "48.24"
        }, {
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "198706",
            "placement": "",
            "impressions": "55509.0",
            "clicks": "116",
            "reach": "42.49"
        }])

        # a file with a line repeated, it is expected to be deduplicated
        self.duplicate_dcm = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": "48988.5",
            "clicks": "118",
            "reach": "48.24"
        }, {
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": "48988.5",
            "clicks": "118",
            "reach": "48.24"
        }])

        # a file with same dimensions values and different metrics values
        self.bad_dcm = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": "48988.5",
            "clicks": "118",
            "reach": "48.24"
        }, {
            "date": "2018-01-01",
            "campaign_id": "86394",
            "campaign": "acme_asprin",
            "placement_id": "267821",
            "placement": "mediamath_programmatic",
            "impressions": "648712.9",
            "clicks": "876",
            "reach": "87.54"
        }])

        # first two lines of a good dsp's file
        self.good_dsp = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "acme_asprin_youtube",
            "impressions": "6011070",
            "clicks": "11889",
            "cost": "40334.2797",
        }, {
            "date": "2018-01-01",
            "campaign_id": "111493",
            "campaign": "acme_car_youtube",
            "impressions": "6585720",
            "clicks": "29843",
            "cost": "58547.0508",
        }])

        # a file with a line repeated, it is expected to be deduplicated
        self.duplicate_dsp = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "acme_asprin_youtube",
            "impressions": "6011070",
            "clicks": "11889",
            "cost": "40334.2797",
        }, {
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "acme_asprin_youtube",
            "impressions": "6011070",
            "clicks": "11889",
            "cost": "40334.2797",
        }])

        # first two line of a file with data missing, will be filled with zero
        self.missing_fine_dsp = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "acme_asprin_youtube",
            "impressions": "6011070",
            "clicks": "",
            "cost": "40334.2797",
        }, {
            "date": "2018-01-01",
            "campaign_id": "111493",
            "campaign": "acme_car_youtube",
            "impressions": "",
            "clicks": "29843",
            "cost": "",
        }])

        # first two line of a file with data missing, in this case, dimensions
        # are missing
        self.missing_bad_dsp = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "",
            "impressions": "6011070",
            "clicks": "11889",
            "cost": "40334.2797",
        }, {
            "date": "2018-01-01",
            "campaign_id": "111493",
            "campaign": "acme_car_youtube",
            "impressions": "6585720",
            "clicks": "29843",
            "cost": "58547.0508",
        }])

        # a file with same dimensions values and different metrics values
        self.bad_dsp = pd.DataFrame([{
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "acme_asprin_youtube",
            "impressions": "6011070",
            "clicks": "11889",
            "cost": "40334.2797",
        }, {
            "date": "2018-01-01",
            "campaign_id": "128115",
            "campaign": "acme_asprin_youtube",
            "impressions": "6585720",
            "clicks": "29843",
            "cost": "58547.0508",
        }])

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_extract(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = pd.DataFrame([{'a': 1, 'b': 3}])

        worker = Worker()
        worker.extract('^dcm.*')
        mock_get_file.assert_called_with('dcm.csv')

        worker = Worker()
        worker.extract('^dbm.*')
        mock_get_file.assert_called_with('dbm.csv')

        worker = Worker()
        worker.extract('^mediamath.*')
        mock_get_file.assert_called_with('mediamath.csv')

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dcm_good(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.good_dcm

        worker = DcmWorker()
        worker.extract('^dcm.*')
        worker.parse()
        self.assertTrue(len(worker.dfs) == 1, "There should be one DataFrame!")

        parsed_df = worker.dfs[0]
        columns_has = parsed_df.columns
        columns_should = [
            'date',
            'campaign_id',
            'campaign',
            'placement_id',
            'placement',
            'impressions',
            'clicks',
            'reach'
        ]

        self.assertListEqual(sorted(columns_has), sorted(columns_should), """
            Missing some column! They should be:
                - date
                - campaign_id
                - campaign
                - placement_id
                - placement
                - impressions
                - clicks
                - reach
            """)

        self.assertEqual(parsed_df.date.dtype, np.dtype('datetime64[ns]'),
                         """date should be parsed to np.datetime64""")
        self.assertEqual(parsed_df.campaign_id.dtype, np.dtype('int64'),
                         """campaign_id should be parsed to np.int64""")
        self.assertEqual(parsed_df.placement_id.dtype, np.dtype('int64'),
                         """placement_id should be parsed to np.int64""")
        self.assertEqual(parsed_df.impressions.dtype, np.dtype('float64'),
                         """impressions should be parsed to np.float64""")
        self.assertEqual(parsed_df.clicks.dtype, np.dtype('int64'),
                         """clicks should be parsed to np.int64""")
        self.assertEqual(parsed_df.reach.dtype, np.dtype('float64'),
                         """reach should be parsed to np.float64""")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dcm_missing(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.missing_fine_dcm
        worker = DcmWorker()
        worker.extract()
        worker.parse()
        self.assertTrue(len(worker.dfs) == 1, "There should be on DataFrame!")
        self.assertTrue(worker.dfs[0].loc[0].clicks ==
                        0, "Clicks should be zero")
        self.assertTrue(worker.dfs[0].loc[0].reach ==
                        0, "Reach should be zero")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dcm_missing_bad(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.missing_bad_dcm
        worker = DcmWorker()
        worker.extract()
        with self.assertRaises(Exception):
            worker.parse()

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dcm_duplicate(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.duplicate_dcm
        worker = DcmWorker()
        worker.extract()
        worker.parse()
        self.assertTrue(worker.dfs[0].shape[0] == 1,
                        "The duplicate rows should be removed!")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dcm_bad_values(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.bad_dcm
        worker = DcmWorker()
        worker.extract()

        # very bad approach...
        working = True
        try:
            worker.parse()
            working = False
        except Exception:
            pass

        if not working:
            self.assertFalse(True, """It should raise an exception since we
                are feeding the worker with a bad dataframe. This dataframe
                has rows with the same dimensions' combination and some
                difference in the metrics values, this combination should be
                unique""")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dsp_good(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.good_dsp

        worker = DspWorker('dbm')
        worker.extract()
        worker.parse()
        self.assertTrue(len(worker.dfs) == 1, "There should be one DataFrame!")

        parsed_df = worker.dfs[0]
        columns_has = parsed_df.columns
        columns_should = [
            'date',
            'campaign_id',
            'campaign',
            'impressions',
            'clicks',
            'cost'
        ]

        self.assertListEqual(sorted(columns_has), sorted(columns_should), """
            Missing some column! They should be:
                - date
                - campaign_id
                - campaign
                - impressions
                - clicks
                - cost
            """)

        self.assertEqual(parsed_df.date.dtype, np.dtype('datetime64[ns]'),
                         """date should be parsed to np.datetime64""")
        self.assertEqual(parsed_df.campaign_id.dtype, np.dtype('int64'),
                         """campaign_id should be parsed to np.int64""")
        self.assertEqual(parsed_df.impressions.dtype, np.dtype('float64'),
                         """impressions should be parsed to np.float64""")
        self.assertEqual(parsed_df.clicks.dtype, np.dtype('int64'),
                         """clicks should be parsed to np.int64""")
        self.assertEqual(parsed_df.cost.dtype, np.dtype('float64'),
                         """reach should be parsed to np.float64""")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dsp_missing(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.missing_fine_dsp
        worker = DspWorker('dbm')
        worker.extract()
        worker.parse()
        self.assertEqual(len(worker.dfs), 1, "There should be on DataFrame!")
        self.assertEqual(worker.dfs[0].loc[0].clicks, 0,
                         "Clicks should be zero")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dsp_missing_bad(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.missing_bad_dsp
        worker = DspWorker('dbm')
        worker.extract()
        with self.assertRaises(Exception):
            worker.parse()

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dsp_duplicate(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.duplicate_dsp
        worker = DspWorker('dbm')
        worker.extract()
        worker.parse()
        self.assertEqual(worker.dfs[0].shape[0], 1,
                         "The duplicate rows should be removed!")

    @patch.object(BucketHelper, 'list_files')
    @patch.object(BucketHelper, 'get_csv_file')
    def test_parse_dsp_bad_values(self, mock_get_file, mock_list_files):
        mock_list_files.return_value = self.fake_list
        mock_get_file.return_value = self.bad_dsp
        worker = DspWorker('dbm')
        worker.extract()

        with self.assertRaises(Exception):
            worker.parse()



class Config(object):
    """
    Common configurations
    """

    DEBUG = True


class DevelopmentConfig(Config):
    """
    Development configurations
    """

    SQLALCHEMY_ECHO = True


class ProductionConfig(Config):
    """
    Production configurations
    """

    DEBUG = False


class TestingConfig(Config):
    """
    Testing configurations
    """

    TESTING = True


app_config = {
    'development': DevelopmentConfig,
    'production': ProductionConfig,
    'testing': TestingConfig
}



# -*- coding: utf-8 -*-

# python standard
import re
import logging

# third-party imports
import pandas as pd
from sqlalchemy import and_

# local imports
from utils.bucket_helper import BucketHelper
from utils.sql_helper import get_connection, get_context
from webapp.app.models import Classification
from webapp.app.queries import GENERATE_REPORT

############################################################################
logger = logging.getLogger('dspreview_application')
con = get_connection()
context = get_context()
############################################################################


class Worker(object):
    """
    This class provides capability for downloading .csv files from a bucket
    and save it to the sql database.
    It classifies their information regarded to brand, sub brand, and dsp.
    """

    def __init__(self):
        self.dfs = []
        self.dfs_classified = []
        self.pattern = None
        self.dsp = None
        self.load_classifications()

    def extract(self, pattern=None):
        """
        this function checks whether there are files in the GCP bucket
        with the given `pattern`. If there are, those files will be
        downloaded

        Params
        ------
        pattern : string
            it is a regex patter for searching in the bucket's files, in case
            of none, it will try to use a pre configured patter, but if it
            was not set rpeviouly, an exception will raise

        Returns
        -------
        The object instace for use in chain calls
        """
        bucket = BucketHelper()
        pattern = pattern or self.pattern
        self.dfs = []
        files = bucket.list_files()
        for f in files:
            fname = f['name']
            if re.search(pattern, fname):
                logger.info("Extracting file [{}]".format(fname))
                self.dfs.append(bucket.get_csv_file(fname))
        return self

    def transform(self):
        """
        Execute all transformation that apply

        Returns
        -------
        The object instace for use in chain calls
        """
        return self.parse().classify()

    def load(self):
        """
        Save data to database

        Returns
        -------
        The object instace for use in chain calls
        """
        return self.upload(raw=True).upload()

    def upload(self, raw=False):
        """this is a basic upload to the mysql database, since the schema
        is defined in src.utils.sql_helper, the DataFrames are expect to be
        in a very specific format, which must be assured by the parsers
        implementation

        Params
        ------
        raw : boolean
            if True, it will upload only the raw data, without classifications

        Returns
        -------
        The object instace for use in chain calls
        """
        if raw:
            logmsg = "Uploading [{}] [raw]".format(self.dsp or "DCM")
            dims = self.dimensions_raw
            table_type = 'raw'
            dfs = self.dfs
        else:
            logmsg = "Uploading [{}] [classified]".format(self.dsp or "DCM")
            dims = self.dimensions
            table_type = 'classified'
            dfs = self.dfs_classified

        logger.info(logmsg)
        table = "{}_{}".format('dsp' if self.dsp else 'dcm', table_type)
        table_temp = "{}_temp".format(table)

        for df in dfs:
            df.to_sql(con=con, name=table_temp,
                      if_exists='replace', index=False)
            update_part = []
            for c in self.metrics:
                update_part.append("{table}.{fld}={temp}.{fld}".format(
                    table=table, temp=table_temp, fld=c))
            update_part.append(
                "{table}.updated_at=CURRENT_TIMESTAMP()".format(table=table))
            update_part = ",".join(update_part)
            all_columns = ",".join(dims+self.metrics)
            connection = con.connect()
            connection.execute("""INSERT INTO {table} ({all_cols})
                                    SELECT {all_cols}
                                    FROM {temp} ON DUPLICATE KEY
                                    UPDATE
                                    {updates}
                                    """.format(table=table, temp=table_temp,
                                               updates=update_part,
                                               all_cols=all_columns))
            connection.execute("DROP TABLE {temp}".format(temp=table_temp))
        return self

    def parse(self):
        raise NotImplementedError("""Implemented by children, for specific
                                  purposes.""")

    def find_by_patter(self, ad_line, classifiers, classifying):
        for c in classifiers:
            composed = ""
            if c.use_campaign_id:
                composed += str(ad_line.get("campaign_id") or 0)
            if c.use_campaign:
                composed += ad_line.get("campaign") or ""
            if c.use_placement_id:
                composed += str(ad_line.get("placement_id") or 0)
            if c.use_placement:
                composed += ad_line.get("placement") or ""
            if re.search(c.pattern, composed, re.IGNORECASE):
                return getattr(c, classifying)
        return "unidentified " + classifying.replace("_", "")

    def find_brand(self, ad_line):
        return self.find_by_patter(ad_line, self.for_brand, "brand")

    def find_sub_brand(self, ad_line):
        return self.find_by_patter(ad_line, self.for_sub_brand, "sub_brand")

    def find_dsp(self, ad_line):
        return self.find_by_patter(ad_line, self.for_dsp, "dsp")

    def classify(self):
        """
        try to classify brand, sub brand, and dsp according to fields values
        it should not be called before parse!

        Returns
        -------
        The object instace for use in chain calls
        """
        logger.info("Applying classification")
        self.dfs_classified = []
        for i, df in enumerate(self.dfs):
            df = df.assign(brand=df.apply(self.find_brand, axis=1)) \
                .assign(sub_brand=df.apply(self.find_sub_brand, axis=1)) \
                .assign(dsp=df.apply(self.find_dsp, axis=1))

            df = df.groupby(self.dimensions) \
                .agg(self.metrics_agg).reset_index()
            self.dfs_classified.append(df.copy())
        return self

    def load_classifications(self):
        """
        Load the classifications for figuring out the brand, sub brand, and
        dsp according to the information in campaign and placement fields
        """
        with context:
            self.for_brand = Classification.query \
                .filter(and_(
                            Classification.brand.isnot(None),
                            Classification.brand != "")).all()
            self.for_sub_brand = Classification.query \
                .filter(and_(
                            Classification.sub_brand.isnot(None),
                            Classification.sub_brand != "")).all()
            self.for_dsp = Classification.query \
                .filter(and_(
                            Classification.dsp.isnot(None),
                            Classification.dsp != "")).all()


class DcmWorker(Worker):
    def __init__(self):
        super(DcmWorker, self).__init__()
        self.pattern = ".*dcm.*"
        self.dimensions_raw = [
            "date",
            "campaign_id",
            "campaign",
            "placement_id",
            "placement"
        ]
        self.dimensions = [
            "date",
            "brand",
            "sub_brand",
            "campaign_id",
            "campaign",
            "placement_id",
            "placement",
            "dsp"
        ]
        self.metrics_agg = {
            "impressions": "sum",
            "clicks": "sum",
            "reach": "sum"  # WARNING! CALCULATED METRIC!!!
        }
        self.metrics = list(self.metrics_agg.keys())

    def parse(self):
        """this function should apply all the proper parsing operations in the
        downloaded dataframes, setting values as dates, integers, or floats

        Returns
        -------
        The object instace for use in chain calls
        """
        logger.info("Parsing DCM file")
        for i, df in enumerate(self.dfs):
            df.date = pd.to_datetime(df.date)
            df.campaign_id = df.campaign_id.replace(
                '', '0', regex=True).fillna(0).astype(int)
            df.placement_id = df.placement_id.replace(
                '', '0', regex=True).fillna(0).astype(int)
            df.impressions = df.impressions.replace(
                '', '0', regex=True).fillna(0).astype(float)
            df.clicks = df.clicks.replace(
                '', '0', regex=True).fillna(0).astype(int)
            df.reach = df.reach.replace(
                '', '0', regex=True).fillna(0).astype(float)
            df.drop_duplicates(inplace=True)

            prob_df = df.groupby(self.dimensions_raw)\
                .agg(self.metrics_agg).reset_index()

            if prob_df.shape[0] != df.shape[0]:
                raise Exception("""Problematic rows!
                    The dimensions combinations should be unique
                    It is not a problem of suplicate rows, it is a problem
                    of some combination having different values
                    of impressions, clicks, or cost.""")

            if not all(df.campaign.values) or not all(df.placement.values):
                raise Exception("""There are missing values in campaign or
                                placement fields""")

            self.dfs[i] = df.copy()
        return self


class DspWorker(Worker):
    def __init__(self, dsp):
        """
        Params
        ------
        dsp : string
            a DSP name, withou formating, spaces or special chars. this will
            be used as pattern for searching advertisenments of this DSP in the
            csv files
        """
        super(DspWorker, self).__init__()
        self.dsp = dsp
        self.pattern = ".*%s.*" % dsp
        self.dimensions_raw = [
            "date",
            "campaign",
            "campaign_id",
        ]
        self.dimensions = [
            "date",
            "brand",
            "sub_brand",
            "dsp",
            "campaign",
            "campaign_id"
        ]
        self.metrics_agg = {
            "impressions": "sum",
            "clicks": "sum",
            "cost": "sum"
        }
        self.metrics = list(self.metrics_agg.keys())

    def parse(self):
        """this function should apply all the proper parsing operations in the
        downloaded dataframes, setting values as dates, integers, or floats

        Returns
        -------
        The object instace for use in chain calls
        """
        logger.info("Parsing DSP file [{}]".format(self.dsp))
        for i, df in enumerate(self.dfs):
            df.date = pd.to_datetime(df.date)
            df.campaign_id = df.campaign_id.replace(
                '', '0', regex=True).fillna(0).astype(int)
            df.impressions = df.impressions.replace(
                '', '0', regex=True).fillna(0).astype(float)
            df.clicks = df.clicks.replace(
                '', '0', regex=True).fillna(0).astype(int)
            df.cost = df.cost.replace(
                '', '0', regex=True).fillna(0).astype(float)
            df.drop_duplicates(inplace=True)

            prob_df = df.groupby(self.dimensions_raw)\
                .agg(self.metrics_agg).reset_index()

            if prob_df.shape[0] != df.shape[0]:
                raise Exception("""Problematic rows!
                    The dimensions combinations should be unique
                    It is not a problem of suplicate rows, it is a problem
                    of some combination having different values
                    of impressions, clicks, or cost.""")

            if not all(df.campaign.values):
                raise Exception("There are missing values in campaign field")

            self.dfs[i] = df.copy()

        return self


def generate_report():
    """ this function generates the full report, joining data from DCM and
    the DSPs. Since the files might have arbitrary dates, the option is to
    generate the whole report table again.
    """
    con = get_connection()
    connection = con.connect()
    connection.execute(GENERATE_REPORT)



# -*- coding: utf-8 -*-

# python standard
import os
import unittest
from unittest.mock import patch
import logging

# local imports
from utils.sql_helper import rand_word, get_connection_strs
from utils.config_helper import ConfigHelper

logging.disable(logging.CRITICAL)


class TestSqlHelper(unittest.TestCase):

    def test_rand_word(self):
        self.assertEqual(len(rand_word(20)), 20, "Error in secret size")

    @patch.dict(os.environ, {
        "DB_HOST": "v1",
        "DB_PORT": "v2",
        "DB_USER": "v3",
        "DB_PASS": "v4",
        "DB_NAME": "v5",
        "DB_TEST_NAME": "v6",
        "DB_TEST_USER": "v7",
        "DB_TEST_PASS": "v8"
    })
    def test_connection_strs_normal(self):
        strs = get_connection_strs()
        full = "mysql+mysqldb://v3:v4@v1:v2/v5"
        simple = "mysql://v3:v4@v1:v2/v5"
        simple_test = "mysql://v7:v8@v1:v2/v6"
        self.assertEqual(strs.con_str, full,
                         "Error in main connection string")
        self.assertEqual(strs.flask_str, simple,
                         "Error in flask connection string")
        self.assertEqual(strs.test_str, simple_test,
                         "Error in test connection string")

    @patch.dict(os.environ, {
        "DB_HOST": "",
        "DB_PORT": "",
        "DB_USER": "",
        "DB_PASS": "",
        "DB_NAME": "",
        "DB_TEST_NAME": "",
        "DB_TEST_USER": "",
        "DB_TEST_PASS": ""
    })
    @patch.object(ConfigHelper, 'get_config')
    def test_search_in_file_valid(self, mock_get_config):
        mock_get_config.return_value = "sv"
        strs = get_connection_strs()
        full = "mysql+mysqldb://sv:sv@sv:sv/sv"
        self.assertEqual(strs.con_str, full,
                         "Error in main connection string")

    @patch.dict(os.environ, {
        "DB_HOST": "",
        "DB_PORT": "",
        "DB_USER": "",
        "DB_PASS": "",
        "DB_NAME": "",
        "DB_TEST_NAME": "",
        "DB_TEST_USER": "",
        "DB_TEST_PASS": ""
    })
    @patch.object(ConfigHelper, 'get_config')
    def test_search_in_file_invalid(self, mock_get_config):
        mock_get_config.return_value = ""
        with self.assertRaises(Exception):
            get_connection_strs()


if __name__ == "__main__":
    unittest.main()



# third-party imports
from flask import flash, redirect, render_template, url_for
from flask_login import login_required, login_user, logout_user

# local imports
from . import auth
from .forms import LoginForm, RegistrationForm
from .. import db
from ..models import User


@auth.route('/register', methods=['GET', 'POST'])
def register():
    """
    Handle requests to the /register route
    Add an user to the database through the registration form
    """
    form = RegistrationForm()
    if form.validate_on_submit():
        user = User(email=form.email.data,
                    username=form.username.data,
                    first_name=form.first_name.data,
                    last_name=form.last_name.data,
                    password=form.password.data)

        # add user to the database
        db.session.add(user)
        db.session.commit()
        flash('You have successfully registered! You may now login.')

        # redirect to the login page
        return redirect(url_for('auth.login'))

    # load registration template
    return render_template('auth/register.html', form=form, title='Register')


@auth.route('/login', methods=['GET', 'POST'])
def login():
    """
    Handle requests to the /login route
    Log an user in through the login form
    """
    form = LoginForm()
    if form.validate_on_submit():

        # check whether user exists in the database and whether
        # the password entered matches the password in the database
        user = User.query.filter_by(email=form.email.data).first()
        if user is not None and user.verify_password(
                form.password.data):
            # log user in
            login_user(user)

            # redirect to the dashboard page after login
            return redirect(url_for('home.dashboard'))

        # when login details are incorrect
        else:
            flash('Invalid email or password.')

    # load login template
    return render_template('auth/login.html', form=form, title='Login')


@auth.route('/logout')
@login_required
def logout():
    """
    Handle requests to the /logout route
    Log an user out through the logout link
    """
    logout_user()
    flash('You have successfully been logged out.')

    # redirect to the login page
    return redirect(url_for('auth.login'))



# python standard
import os

# local imports
from webapp.app import create_app

config_name = os.getenv('FLASK_CONFIG') or "production"
app = create_app(config_name)

if __name__ == '__main__':
    app.run()



# -*- coding: utf-8 -*-

# python standard
import sys
import os
import re
import string
import logging
from collections import namedtuple
from random import choices

# third-party imports
from sqlalchemy import create_engine

# local imports
from utils.config_helper import ConfigHelper

############################################################################
logger = logging.getLogger('dspreview_application')
############################################################################


def rand_word(N):
    """
    generate a secret of size N
    """
    lowers = string.ascii_lowercase
    uppers = string.ascii_uppercase
    digits = string.digits
    return ''.join(choices(lowers + uppers + digits, k=N))


def get_connection_info():
    ci = namedtuple("ConnectionInfo",
                    "dbhost, dbport, dbuser, dbpass, dbname")
    config = ConfigHelper()
    ci.dbhost = os.getenv("DB_HOST") or config.get_config("DB_HOST")
    ci.dbport = os.getenv("DB_PORT") or config.get_config("DB_PORT")
    ci.dbuser = os.getenv("DB_USER") or config.get_config("DB_USER")
    ci.dbpass = os.getenv("DB_PASS") or config.get_config("DB_PASS")
    ci.dbname = os.getenv("DB_NAME") or config.get_config("DB_NAME")
    return ci


def get_connection_info_test():
    ci = namedtuple("ConnectionInfoTest",
                    "dbhost, dbport, dbuser, dbpass, dbname")
    config = ConfigHelper()
    ci.dbhost = os.getenv("DB_HOST") or config.get_config("DB_HOST")
    ci.dbport = os.getenv("DB_PORT") or config.get_config("DB_PORT")
    ci.dbname = os.getenv("DB_TEST_NAME") or config.get_config("DB_TEST_NAME")
    ci.dbuser = os.getenv("DB_TEST_USER") or config.get_config("DB_TEST_USER")
    ci.dbpass = os.getenv("DB_TEST_PASS") or config.get_config("DB_TEST_PASS")
    return ci


def get_connection_strs():
    ci = get_connection_info()
    cit = get_connection_info_test()
    # all these values are necessary
    if not all([ci.dbhost, ci.dbport, ci.dbuser, ci.dbpass, ci.dbname]):
        raise Exception("""It is not possible to connect in the MySQL
                        database. All of the following environment
                        variables must be set
                        - DB_HOST
                        - DB_PORT
                        - DB_USER
                        - DB_PASS
                        - DB_NAME
                        - DB_TEST_NAME (for development only)
                        - DB_TEST_USER (for development only)
                        - DB_TEST_PASS (for development only)
                        another possibility woul be the .dspreview.json
                        ;)""")
    # connection string for operational use
    full = "mysql+mysqldb://{dbuser}:{dbpass}@{dbhost}:{dbport}/{dbname}"
    simple = "mysql://{dbuser}:{dbpass}@{dbhost}:{dbport}/{dbname}"

    constr = namedtuple("ConnectionStrings",
                        "con_str, flask_str, test_str")
    constr.con_str = full.format(dbuser=ci.dbuser, dbpass=ci.dbpass,
                                 dbhost=ci.dbhost, dbport=ci.dbport,
                                 dbname=ci.dbname)
    constr.flask_str = simple.format(dbuser=ci.dbuser, dbpass=ci.dbpass,
                                     dbhost=ci.dbhost, dbport=ci.dbport,
                                     dbname=ci.dbname)
    constr.test_str = simple.format(dbuser=cit.dbuser,
                                    dbpass=cit.dbpass,
                                    dbhost=cit.dbhost, dbport=cit.dbport,
                                    dbname=cit.dbname)
    return constr


def initialize_database():
    try:
        flask_str = get_connection_strs().flask_str

        # CREATE CONFIG FILES
        logger.info("Creating configuration files")
        instance_folder = sys.prefix + "/var/webapp.app-instance"
        if not os.path.exists(instance_folder):
            os.makedirs(instance_folder)
        config_file = instance_folder + "/config.py"
        config_lines = []

        track_mod_line = "SQLALCHEMY_TRACK_MODIFICATIONS=False\n"
        sql_config_line = "SQLALCHEMY_DATABASE_URI='{constr}'\n"
        sql_config_line = sql_config_line.format(constr=flask_str)
        secret_key_line = "SECRET_KEY='{secret}'\n"
        secret_key_line = secret_key_line.format(secret=rand_word(20))

        if os.path.exists(config_file):
            with open(config_file, "r") as f:
                config_lines = f.readlines()

        has_uri_cfg = False
        has_secret = False
        has_track_mod = False

        for i in range(len(config_lines)):
            if re.search(".*SQLALCHEMY_DATABASE_URI.*", config_lines[i]):
                config_lines[i] = sql_config_line
                has_uri_cfg = True
            if re.search(".*SECRET_KEY.*", config_lines[i]):
                has_secret = True
            if re.search(".*SQLALCHEMY_TRACK_MODIFICATIONS.*",
                         config_lines[i]):
                has_track_mod = True

        if not has_uri_cfg:
            config_lines.append(sql_config_line)
        if not has_secret:
            config_lines.append(secret_key_line)
        if not has_track_mod:
            config_lines.append(track_mod_line)

        with open(config_file, "w") as f:
            for l in config_lines:
                f.write(l)

        # CREATE CONFIG FILES
        logger.info("Configuring database")
        from webapp.app import db, create_app
        app = create_app('production')
        app.config.update(
            SQLALCHEMY_DATABASE_URI=flask_str
        )
        with app.app_context():
            db.create_all()
    except Exception as err:
        logger.exception(err)


def get_context():
    """
    This context is necessary for using the flask models outside the app

    example:
        > with get_context():
        >     cls = Classifications.query.all()
    """
    flask_str = get_connection_strs().flask_str
    logger.info("Creating app context")
    from webapp.app import create_app
    app = create_app('production')
    app.config.update(
        SQLALCHEMY_DATABASE_URI=flask_str
    )
    return app.app_context()


def get_connection():
    """
    Create a connection to the MySQL database
    """
    con_str = get_connection_strs().con_str
    logger.info("Creating sql connection")
    return create_engine(con_str, pool_recycle=1, pool_timeout=57600).connect()



# -*- coding: utf-8 -*-

# python standard
import sys
from datetime import datetime
sys.path.insert(0, "./src")
sys.path.insert(0, "./src/webapp")

# third-party imports
from flask_testing import TestCase
from flask import abort, url_for

# local imports
from utils.sql_helper import get_connection_strs
from webapp.app import create_app, db
from webapp.app.models import User, Classification
from webapp.app.models import DCMRaw, DCM, DSPRaw, DSP, Report


class TestBase(TestCase):

    def create_app(self):
        # pass in test configurations
        config_name = 'testing'
        app = create_app(config_name)
        app.config.update(
            SQLALCHEMY_DATABASE_URI=get_connection_strs().test_str
        )

        return app

    def setUp(self):
        # it needs to be here
        """
        Will be called before every test
        """

        db.create_all()

        # create test admin user
        admin = User(username="admin", password="admin2018", is_admin=True)

        # create test non-admin user
        normal_user = User(username="normal_user", password="test2018")

        # create test for classification
        classification = Classification(pattern=".*", brand="some brand",
                                        sub_brand="some sub brand",
                                        dsp="some dsp",
                                        use_campaign_id=True,
                                        use_campaign=False,
                                        use_placement_id=False,
                                        use_placement=True)

        # create a test for dcm
        dcm = DCM(date=datetime.now(), campaign_id=85989,
                  campaign="some campaign", placement_id=54786,
                  placement="some placement", impressions=87884.8,
                  clicks=874, reach=7581.5, brand="some brand",
                  sub_brand="some sub brand", dsp="some dsp")

        # create a test for dcm raw
        dcm_raw = DCMRaw(date=datetime.now(), campaign_id=85989,
                         campaign="some campaign", placement_id=54786,
                         placement="some placement", impressions=87884.8,
                         clicks=874, reach=7581.5)

        # create a test for dsp
        dsp = DSP(date=datetime.now(), campaign_id=85989,
                  campaign="some campaign", impressions=87884.8,
                  clicks=874, cost=7581.5, brand="some brand",
                  sub_brand="some sub brand", dsp="some dsp")

        # create a test for dsp raw
        dsp_raw = DSPRaw(date=datetime.now(), campaign_id=85989,
                         campaign="some campaign", impressions=87884.8,
                         clicks=874, cost=7581.5)

        report = Report(date=datetime.now(), brand="some brand",
                        sub_brand="some sub brand", ad_campaign_id=89865,
                        ad_campaign="some campaign", dsp="some dsp",
                        dsp_campaign_id=87897, dsp_campaign="some campaign",
                        ad_impressions=783546.8, ad_clicks=578698,
                        ad_reach=879.9, dsp_impressions=7987.9,
                        dsp_clicks=7897, dsp_cost=578979.8)

        # save users to database
        db.session.add(admin)
        db.session.add(normal_user)
        db.session.add(classification)
        db.session.add(dcm)
        db.session.add(dcm_raw)
        db.session.add(dsp)
        db.session.add(dsp_raw)
        db.session.add(report)
        db.session.commit()

    def tearDown(self):
        """
        Will be called after every test
        """

        db.session.remove()
        db.drop_all()


class TestModels(TestBase):

    def test_user_model(self):
        """
        Test number of records in User table
        """
        self.assertEqual(User.query.count(), 2)

    def test_classification_model(self):
        """
        Test number of records in Classification table
        """
        self.assertEqual(Classification.query.count(), 1)

    def test_dsp_model(self):
        """
        Test number of records in DSP table
        """
        self.assertEqual(DSP.query.count(), 1)

    def test_dsp_raw_model(self):
        """
        Test number of records in DSPRaw table
        """
        self.assertEqual(DSPRaw.query.count(), 1)

    def test_dcm_model(self):
        """
        Test number of records in DCM table
        """
        self.assertEqual(DCM.query.count(), 1)

    def test_dcm_raw_model(self):
        """
        Test number of records in DCMRaw table
        """
        self.assertEqual(DCMRaw.query.count(), 1)

    def test_report_model(self):
        """
        Test number of records in DCMRaw table
        """
        self.assertEqual(Report.query.count(), 1)



# -*- coding: utf-8 -*-

# python standard
import os
import unittest
import unittest.mock as mock
import logging
from unittest.mock import patch, mock_open

# local imports
from utils.bucket_helper import BucketHelper
from utils.config_helper import ConfigHelper

logging.disable(logging.CRITICAL)


class TestBucketHelper(unittest.TestCase):

    @mock.patch.dict(os.environ, {
        "GCP_BUCKET": "some_value",
        "GOOGLE_APPLICATION_CREDENTIALS": "some_value"
    })
    def test_instance_with_envs(self):
        bh = BucketHelper()
        self.assertIsInstance(bh, BucketHelper)

    @mock.patch.dict(os.environ, {
        "GCP_BUCKET": '',
        "GOOGLE_APPLICATION_CREDENTIALS": ''
    })
    @mock.patch('utils.bucket_helper.os.path.exists')
    def test_instance_with_valid_file(self, mock_path_exists):
        mock_path_exists.return_value = True
        read_data = """{
                    "GOOGLE_APPLICATION_CREDENTIALS": "some_value",
                    "GCP_BUCKET": "dspreview"
                    }"""
        mo = mock_open(read_data=read_data)
        with patch('utils.bucket_helper.open', mo):
            bh = BucketHelper()
            self.assertIsInstance(bh, BucketHelper)

    @mock.patch.dict(os.environ, {
        "GCP_BUCKET": '',
        "GOOGLE_APPLICATION_CREDENTIALS": ''
    })
    @patch.object(ConfigHelper, 'get_config')
    def test_instance_with_invalid_file(self, mock_config):
        mock_config.return_value = None

        # very bad approach...
        working = True
        try:
            BucketHelper()
            working = False
        except Exception:
            pass

        if not working:
            self.assertFalse(True, """It should raise an exception since
            we are feeding a invalid configuration""")



# -*- coding: utf-8 -*-
"""
Config file manager
"""

# python standard
import os
import json


class ConfigHelper(object):
    """
    Manage the config file
    """
    def __init__(self):
        self.config_info = {}

        # check for the config file even if the env vars are found
        self.user_home = os.path.expanduser("~")
        self.config_file = "{}/.dspreview.json".format(self.user_home)

        # create a config file if it does not exist
        if not os.path.exists(self.config_file):
            with open(self.config_file, 'a') as f:
                f.write("")
        else:
            with open(self.config_file, 'r') as f:
                try:
                    self.config_info = json.loads(f.read())
                except Exception as err:
                    print(str(err))
                    raise Exception("Invalid .dspreview.json file.")

    def get_config(self, name):
        """
        Params
        -----
        name : string
            the nome of the parameter to be searched in the config file
        """
        return self.config_info.get(name)

    def set_config(self, name, value):
        self.config_info[name] = value
        content = json.dumps(self.config_info)
        with open(self.config_file, 'w') as f:
            f.write(content)

        # reload contents
        self.__init__()



# -*- coding: utf-8 -*-

from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField, BooleanField
from wtforms.validators import DataRequired


class ClassificationForm(FlaskForm):
    """
    Form for adding, editing, removing classifications
    """
    pattern = StringField('Pattern', validators=[DataRequired()],
                          render_kw={"placeholder": "Regular Expression..."})
    brand = StringField('Brand', render_kw={"placeholder": "Brand..."})
    sub_brand = StringField('Sub Brand',
                            render_kw={"placeholder": "Sub Brand..."})
    dsp = StringField('DSP', render_kw={"placeholder": "DSP..."})
    use_campaign_id = BooleanField('Use Campaign ID')
    use_campaign = BooleanField('Use Campaign')
    use_placement_id = BooleanField('Use Placement ID (DCM)')
    use_placement = BooleanField('Use Placement (DCM)')
    submit = SubmitField('Submit')



# -*- coding: utf-8 -*-
"""
"""

# python standard
import os
import unittest
import logging

# local imports
from workers import cli

logging.disable(logging.CRITICAL)


class TestCli(unittest.TestCase):
    def setUp(self):
        os.environ["FLASK_CONFIG"] = "testing"
        self.parser = cli.create_parser()

    def test_init(self):
        """ init does not require other params """
        args = self.parser.parse_args(['init'])
        self.assertTrue(args.action == "init", "Even init is failing...")

    def test_run_worker(self):
        args = self.parser.parse_args(["--worker", "dcm"])
        self.assertTrue(args.action == "work", "Action should be work!")
        self.assertTrue(args.worker == "dcm", "Worker name is wrong")

    def test_run_worker_with_dsp(self):
        args = self.parser.parse_args(["--worker", "dsp", "--dsp", "dbm"])
        self.assertTrue(args.action == "work", "Action should be work!")
        self.assertTrue(args.dsp == "dbm", "DSP name is wrong!")

    def test_run_dsp_without_worker(self):
        args = self.parser.parse_args(["--dsp", "dbm"])
        self.assertTrue(args.action == "work", "Action should be work!")
        self.assertTrue(args.dsp == "dbm", "DSP name is wrong!")
        self.assertTrue(args.worker == "dsp", "It should be a dsp worker!")

    def test_generate_report(self):
        args = self.parser.parse_args(['--generate-report'])
        self.assertTrue(args.action == "work", "Action should be work!")
        self.assertTrue(args.generate_report, "It should be True!")

    def test_server_with_port(self):
        args = self.parser.parse_args(['serve', '--port', '8080'])
        self.assertTrue(args.action == "serve", "Action should be serve!")
        self.assertTrue(type(args.port) is int, "Port must be an integer!")

    def test_server_without_port(self):
        args = self.parser.parse_args(['serve'])
        self.assertTrue(args.action == "serve", "Action should be serve!")
        self.assertTrue(type(args.port) is int, "Port must be an integer!")



# -*- coding: utf-8 -*-
"""
this module has a friedly interface for dealing with buckets
currently it works only for GCP Storage, but it can be
expanded to work with AWS S3 as well
"""

# python standard
import os
import tempfile
import datetime
import re
import logging

# third-party imports
import pandas as pd
import googleapiclient.discovery

# local imports
from utils.config_helper import ConfigHelper

############################################################################
logger = logging.getLogger('dspreview_application')
############################################################################


class BucketHelper(object):
    """
    List, download, and archive .csv files in buckets
    """

    def __init__(self):
        config = ConfigHelper()
        self.bucket = os.environ.get("GCP_BUCKET") or \
            config.get_config("GCP_BUCKET")
        self.account = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS") or \
            config.get_config("GOOGLE_APPLICATION_CREDENTIALS")

        # all these values are necessary
        if not self.bucket or not self.account:
            raise Exception("""The GCP_BUCKET or GOOGLE_APPLICATION_CREDENTIALS
            are not set and there is not a .dspreview.json file in the user's
            home folder, please provide one of them.""")

        # make sure they are the same
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.account
        self._service = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """WARNING! it needs to be properly closed... """
        self._service = None

    @property
    def service(self):
        """it avoids the creation of multiple services"""
        if not self._service:
            logger.info("Creating Google API client")
            self._service = googleapiclient.discovery.build('storage', 'v1')
        return self._service

    def list_files(self):
        """
        List all files inside the bucket

        Returns
        -------
        files : array_like
           an array of dictionaies like:
           `[{'name': 'dbm.csv', 'contentType': 'text/csv', 'size': '21645'}]`
        """
        fields = 'nextPageToken,items(name,size,contentType,metadata(my-key))'
        logger.info("Listing files in bucket [{}]".format(self.bucket))
        req = self.service.objects().list(bucket=self.bucket, fields=fields)
        all_objects = []
        while req:
            resp = req.execute()
            all_objects.extend(resp.get('items', []))
            req = self.service.objects().list_next(req, resp)
        return all_objects

    def get_csv_file(self, filename):
        """
        Download a file that exists!

        Params
        ------
        filename : string
            the name of a existing file, it might be found through `list_files`

        Returns
        -------
        Pandas dataframe with the parsed .csv file or None if it does not exist
        """
        logger.info("Downloading [{}] from [{}]".format(filename, self.bucket))
        req = self.service.objects().get_media(bucket=self.bucket,
                                               object=filename)
        with tempfile.TemporaryFile(mode='w+b') as tmpfile:
            downloader = googleapiclient.http.MediaIoBaseDownload(tmpfile, req)
            done = False
            while done is False:
                status, done = downloader.next_chunk()
                logger.info("[{}] Download {}%.".format(filename,
                            int(status.progress() * 100)))
            tmpfile.seek(0)
            return pd.read_csv(tmpfile)
        return None

    def archive_csv_file(self, filename):
        """
        Archive a file that exists!

        Params
        ------
        filename : string
            the name of a existing file, it might be found through `list_files`

        Returns
        -------
        A boolean indicating if the file was archived successfully
        """
        try:
            # it will copy and delete for now
            ts = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            new_filename = "archive/{0}___{1}".format(filename, ts)
            req = self.service.objects().copy(sourceBucket=self.bucket,
                                              sourceObject=filename,
                                              destinationBucket=self.bucket,
                                              destinationObject=new_filename,
                                              body={})
            resp = req.execute()
            if resp.get('name') and resp.get('name') == new_filename:
                self.service.objects().delete(bucket=self.bucket,
                                              object=filename).execute()
        except Exception as err:
            logger.info(str(err))
            return False
        return True

    @classmethod
    def dsp_available(cls):
        """
        Guess which files belong to DSP in the bucket

        Returns
        -------
        array of DSPs' files names
        """
        logger.info("Trying to list DSP files")
        with cls() as inst:
            files_list = inst.list_files()
            dsp_files = []
            for f in files_list:
                fname = f['name']
                if len(fname.split('/')) > 1:
                    continue
                elif re.search('.*dcm.*', fname, re.IGNORECASE):
                    continue
                dsp_files.append(fname.split('.')[0])
            return dsp_files



# third-party imports
from flask_wtf import FlaskForm
from wtforms import PasswordField, StringField, SubmitField, ValidationError
from wtforms.validators import DataRequired, Email, EqualTo

# local imports
from ..models import User


class RegistrationForm(FlaskForm):
    """
    Form for users to create new account
    """
    email = StringField('Email', validators=[DataRequired(), Email()])
    username = StringField('Username', validators=[DataRequired()])
    first_name = StringField('First Name', validators=[DataRequired()])
    last_name = StringField('Last Name', validators=[DataRequired()])
    password = PasswordField('Password', validators=[
                                        DataRequired(),
                                        EqualTo('confirm_password')
                                        ])
    confirm_password = PasswordField('Confirm Password')
    submit = SubmitField('Register')

    def validate_email(self, field):
        if User.query.filter_by(email=field.data).first():
            raise ValidationError('Email is already in use.')

    def validate_username(self, field):
        if User.query.filter_by(username=field.data).first():
            raise ValidationError('Username is already in use.')


class LoginForm(FlaskForm):
    """
    Form for users to login
    """
    email = StringField('Email', validators=[DataRequired(), Email()])
    password = PasswordField('Password', validators=[DataRequired()])
    submit = SubmitField('Login')



# -*- coding: utf-8 -*-

# python standard
import unittest
import logging
import unittest.mock as mock
from unittest.mock import patch, mock_open

# local imports
from utils.config_helper import ConfigHelper

logging.disable(logging.CRITICAL)


class TestConfigHelper(unittest.TestCase):

    @mock.patch('utils.config_helper.os.path.exists')
    def test_with_invalid_file(self, mock_path_exists):
        mock_path_exists.return_value = True
        read_data = "{..."
        mo = mock_open(read_data=read_data)
        with patch('utils.config_helper.open', mo):
            with self.assertRaises(Exception):
                ConfigHelper()

    @mock.patch('utils.config_helper.os.path.exists')
    def test_with_valid_file(self, mock_path_exists):
        mock_path_exists.return_value = True
        read_data = """{"key":"value"}"""
        mo = mock_open(read_data=read_data)
        with patch('utils.config_helper.open', mo):
            try:
                ConfigHelper()
            except Exception:
                self.fail("""It shouldn't raise an exception since
                we are feeding a valid file""")

    @mock.patch('utils.config_helper.os.path.exists')
    def test_get_value(self, mock_path_exists):
        mock_path_exists.return_value = True
        read_data = """{"key":"value"}"""
        mo = mock_open(read_data=read_data)
        with patch('utils.config_helper.open', mo):
            config = ConfigHelper()
            self.assertEqual(config.get_config("key"),
                             "value", """Is is not finding the correct
                                        key-value config""")



from setuptools import setup, find_packages

with open('requirements.txt') as fp:
    install_requires = fp.read().split('\n')

with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name='dspreview',
    description='a simple preview for dsp digital advertising information',
    long_description=readme,
    author='Thiago Melo',
    author_email='thiago.lc.melo@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'dspreview=workers.cli:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    url='https://github.com/thiagolcmelo/dspreview',
    download_url='https://github.com/thiagolcmelo/dspreview/archive/v0.1.3.tar.gz',
    version='0.1.3',
)



# -*- coding: utf-8 -*-
GENERATE_CLASSIFIED = """
INSERT INTO
    dcm_classified (date, brand, sub_brand, dsp, campaign_id, campaign,
                    placement, placement_id, impressions, clicks, reach)
    SELECT
        date,
        brand,
        sub_brand,
        dsp,
        campaign_id,
        campaign,
        placement,
        placement_id,
        impressions,
        clicks,
        reach
    FROM
        (
            SELECT
                raw.date,
                ifnull(classified.brand, "unidentified brand") AS brand,
                ifnull(classified.sub_brand, "unidentified sub brand")
                    AS sub_brand,
                ifnull(classified.dsp, "unidentified dsp") AS dsp,
                raw.campaign_id,
                raw.campaign,
                raw.placement_id,
                raw.placement,
                raw.impressions,
                raw.clicks,
                raw.reach
            FROM
                dcm_raw AS raw
                LEFT JOIN
                    (
                        SELECT
                            id,
                            date,
                            campaign_id,
                            campaign,
                            placement_id,
                            placement,
                            MAX(impressions) AS impressions,
                            MAX(clicks) AS clicks,
                            MAX(reach) AS reach,
                            MAX(brand) brand,
                            MAX(sub_brand) AS sub_brand,
                            MAX(dsp) AS dsp
                        FROM
                            (
                                SELECT
                                    dcm.id AS id,
                                    dcm.date AS date,
                                    dcm.campaign_id AS campaign_id,
                                    dcm.campaign AS campaign,
                                    dcm.placement_id AS placement_id,
                                    dcm.placement AS placement,
                                    dcm.impressions AS impressions,
                                    dcm.clicks AS clicks,
                                    dcm.reach AS reach,
                                    concat(
                                        IF(
                                            cls.use_campaign,
                                            dcm.campaign, ''),
                                        IF(
                                            cls.use_campaign_id,
                                            dcm.campaign_id, ''),
                                        IF(
                                            cls.use_placement_id,
                                            dcm.placement_id, ''),
                                        IF(
                                            cls.use_placement,
                                            dcm.placement, '')
                                        ) AS composed,
                                    cls.pattern AS pattern,
                                    cls.brand AS brand,
                                    cls.sub_brand AS sub_brand,
                                    cls.dsp AS dsp
                                FROM
                                    dcm_raw AS dcm
                                    CROSS JOIN
                                        classifications AS cls
                                HAVING
                                    composed REGEXP pattern
                            )
                            AS big_cross
                        GROUP BY
                            id,
                            date,
                            campaign_id,
                            campaign,
                            placement_id,
                            placement
                    )
                    AS classified
                    ON raw.id = classified.id
        )
        AS big_join;
INSERT INTO
    dsp_classified (date, brand, sub_brand, dsp, campaign_id, campaign,
                    impressions, clicks, cost)
    SELECT
        date,
        brand,
        sub_brand,
        dsp,
        campaign_id,
        campaign,
        impressions,
        clicks,
        cost
    FROM
        (
            SELECT
                raw.date,
                ifnull(classified.brand, "unidentified brand") AS brand,
                ifnull(classified.sub_brand, "unidentified sub brand")
                    AS sub_brand,
                ifnull(classified.dsp, "unidentified dsp") AS dsp,
                raw.campaign_id,
                raw.campaign,
                raw.impressions,
                raw.clicks,
                raw.cost
            FROM
                dsp_raw AS raw
                LEFT JOIN
                    (
                        SELECT
                            id,
                            date,
                            campaign_id,
                            campaign,
                            MAX(impressions) AS impressions,
                            MAX(clicks) AS clicks,
                            MAX(cost) AS cost,
                            MAX(brand) brand,
                            MAX(sub_brand) AS sub_brand,
                            MAX(dsp) AS dsp
                        FROM
                            (
                                SELECT
                                    dsp.id AS id,
                                    dsp.date AS date,
                                    dsp.campaign_id AS campaign_id,
                                    dsp.campaign AS campaign,
                                    dsp.impressions AS impressions,
                                    dsp.clicks AS clicks,
                                    dsp.cost AS cost,
                                    concat(
                                        IF(
                                            cls.use_campaign,
                                            dsp.campaign, ''),
                                        IF(
                                            cls.use_campaign_id,
                                            dsp.campaign_id, '')
                                        ) AS composed,
                                    cls.pattern AS pattern,
                                    cls.brand AS brand,
                                    cls.sub_brand AS sub_brand,
                                    cls.dsp AS dsp
                                FROM
                                    dsp_raw AS dsp
                                    CROSS JOIN
                                        classifications AS cls
                                HAVING
                                    composed REGEXP pattern
                            )
                            AS big_cross
                        GROUP BY
                            id,
                            date,
                            campaign_id,
                            campaign
                    )
                    AS classified
                    ON raw.id = classified.id
        )
        AS big_join;
"""

GENERATE_REPORT = """
INSERT INTO
    report (date, brand, sub_brand, ad_campaign_id, ad_campaign,
            dsp_campaign_id, dsp, dsp_campaign, ad_impressions, ad_clicks,
            ad_reach, dsp_impressions, dsp_clicks, dsp_cost)
    SELECT
        date,
        brand,
        sub_brand,
        ad_campaign_id,
        ad_campaign,
        dsp_campaign_id,
        dsp,
        dsp_campaign,
        ad_impressions,
        ad_clicks,
        ad_reach,
        dsp_impressions,
        dsp_clicks,
        dsp_cost
    FROM
        (
            SELECT
                dcm.date AS date,
                dcm.brand AS brand,
                dcm.sub_brand AS sub_brand,
                dcm.campaign_id AS ad_campaign_id,
                dcm.campaign AS ad_campaign,
                dcm.dsp AS dsp,
                dsp.campaign_id AS dsp_campaign_id,
                dsp.campaign AS dsp_campaign,
                SUM(dcm.impressions) AS ad_impressions,
                SUM(dcm.clicks) AS ad_clicks,
                SUM(dcm.reach) AS ad_reach,
                SUM(dsp.impressions) AS dsp_impressions,
                SUM(dsp.clicks) AS dsp_clicks,
                SUM(dsp.cost) AS dsp_cost
            FROM
                dcm_classified AS dcm
                LEFT JOIN
                    dsp_classified AS dsp
                    ON dcm.date = dsp.date
                    AND dcm.brand = dsp.brand
                    AND dcm.sub_brand = dsp.sub_brand
                    AND dcm.dsp = dsp.dsp
            GROUP BY
                dcm.date,
                dcm.brand,
                dcm.sub_brand,
                dcm.campaign_id,
                dcm.campaign,
                dcm.dsp,
                dsp.campaign_id,
                dsp.campaign
        )
        AS big_join
        ON DUPLICATE KEY
        UPdate
            report.ad_impressions = big_join.ad_impressions,
            report.ad_clicks = big_join.ad_clicks,
            report.ad_reach = big_join.ad_reach,
            report.dsp_impressions = big_join.dsp_impressions,
            report.dsp_clicks = big_join.dsp_clicks,
            report.dsp_cost = big_join.dsp_cost,
            report.updated_at = CURRENT_TIMESTAMP();
"""



# python standard
from datetime import datetime

# third-party imports
from flask import flash, redirect, render_template, url_for, jsonify, request
from flask_login import login_required

# local imports
from . import home
from .forms import ClassificationForm
from .. import db
from ..models import Report, Classification, DCM, DSP
from ..queries import GENERATE_CLASSIFIED, GENERATE_REPORT


@home.route('/')
def homepage():
    """
    Render the homepage template on the / route
    """
    return render_template('home/index.html', title="Welcome")


@home.route('/dashboard')
@login_required
def dashboard():
    """
    Render the dashboard template on the /dashboard route
    """
    return render_template('home/dashboard.html', title="Dashboard")


@home.route('/report')
@login_required
def report_date():
    """
    Return full report data as Json
    """
    try:
        def parse_date(dt):
            if dt:
                try:
                    dt = datetime.strptime(dt, "%Y-%m-%d")
                except Exception:
                    dt = None
            return dt

        start_date = parse_date(request.args.get('start_date',
                                                 default=None, type=None))
        end_date = parse_date(request.args.get('end_date',
                                               default=None, type=None))

        results = []
        if start_date and end_date:
            results = Report.query\
                            .filter(Report.date >= start_date)\
                            .filter(Report.date <= end_date)
        elif start_date:
            results = Report.query.filter(Report.date >= start_date)
        elif end_date:
            results = Report.query.filter(Report.date <= end_date)
        else:
            results = Report.query.all()

        return jsonify({
            "status": "success",
            "data": [r.serialize for r in results]
        })
    except Exception as err:
        print(str(err))
        return jsonify({
            "status": "fail",
            "data": []
        })


@home.route('/classifications', methods=['GET', 'POST'])
@login_required
def list_classifications():
    """
    List all classifications
    """
    classifications = Classification.query.all()

    return render_template('home/classifications.html',
                           classifications=classifications,
                           title="Classifications")


@home.route('/classifications/add', methods=['GET', 'POST'])
@login_required
def add_classification():
    """
    Add a classification to the database
    """
    add_classification = True

    form = ClassificationForm()
    if form.validate_on_submit():
        classification = Classification(pattern=form.pattern.data,
                                        brand=form.brand.data,
                                        sub_brand=form.sub_brand.data,
                                        dsp=form.dsp.data,
                                        use_campaign_id=form.use_campaign_id.data,
                                        use_campaign=form.use_campaign.data,
                                        use_placement_id=form.use_placement_id.data,
                                        use_placement=form.use_placement.data)
        try:
            # add classification to the database
            db.session.add(classification)
            db.session.commit()
            flash('You have successfully added a new classification.')
        except Exception:
            # in case classification name already exists
            flash('Error: classification name already exists.')

        # redirect to classifications page
        return redirect(url_for('home.list_classifications'))

    # load classification template
    return render_template('home/classification.html', action="Add",
                           add_classification=add_classification, form=form,
                           title="Add Classification")


@home.route('/classifications/edit/<int:id>', methods=['GET', 'POST'])
@login_required
def edit_classification(id):
    """
    Edit a classification
    """

    add_classification = False

    classification = Classification.query.get_or_404(id)
    form = ClassificationForm(obj=classification)
    if form.validate_on_submit():
        classification.brand = form.brand.data
        classification.sub_brand = form.sub_brand.data
        classification.dsp = form.dsp.data
        classification.use_campaign_id = form.use_campaign_id.data
        classification.use_campaign = form.use_campaign.data
        classification.use_placement_id = form.use_placement_id.data
        classification.use_placement = form.use_placement.data
        db.session.commit()
        flash('You have successfully edited the classification.')

        # redirect to the classifications page
        return redirect(url_for('home.list_classifications'))

    form.brand.data = classification.brand
    form.sub_brand.data = classification.sub_brand
    form.dsp.data = classification.dsp
    form.use_campaign_id.data = classification.use_campaign_id
    form.use_campaign.data = classification.use_campaign
    form.use_placement_id.data = classification.use_placement_id
    form.use_placement.data = classification.use_placement

    return render_template('home/classification.html', action="Edit",
                           add_classification=add_classification, form=form,
                           classification=classification,
                           title="Edit Classification")


@home.route('/classification/delete/<int:id>', methods=['GET', 'POST'])
@login_required
def delete_classification(id):
    """
    Delete a classification from the database
    """
    classification = Classification.query.get_or_404(id)
    db.session.delete(classification)
    db.session.commit()
    flash('You have successfully deleted the classification.')

    # redirect to the classifications page
    return redirect(url_for('home.list_classifications'))

    return render_template(title="Delete Classification")


@home.route('/classification/reset', methods=['GET'])
@login_required
def reset_classifications():
    """
    Reset all classifications, might take a while
    """
    try:
        DCM.query.delete()
        DSP.query.delete()
        Report.query.delete()
        db.session.execute(GENERATE_CLASSIFIED)
        db.session.execute(GENERATE_REPORT)
        db.session.commit()
        return jsonify({
            "status": "success"
        })
    except Exception as err:
        print(str(err))
        return jsonify({
            "status": "fail"
        })



# -*- coding: utf-8 -*-
"""
"""

# python standard
import sys
import time
import argparse
import logging

# local imports
from utils.sql_helper import initialize_database
from utils.bucket_helper import BucketHelper
from workers.worker import DcmWorker, DspWorker, generate_report

############################################################################
str_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(level=logging.DEBUG,
                    format=str_format,
                    datefmt='%Y-%m-%d %H:%M',
                    filename='/tmp/dspreview_application.log',
                    filemode='w')
logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.CRITICAL)
console = logging.StreamHandler()
console.setLevel(logging.INFO)
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
console.setFormatter(formatter)
logging.getLogger('').addHandler(console)
logger = logging.getLogger('dspreview_application')
############################################################################


class ChangeWorker(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.worker = 'dsp'
        setattr(namespace, self.dest, values)


def create_parser():
    parser = argparse.ArgumentParser(description="""
    CLI to control workers execution. It is also useful for initializing
    the database and serving a web app where it is possible to have a
    report.
    """)
    parser.add_argument("action", type=str,
                        help="""It might be 'init for initialize the database
                        or 'serve' for serving the web app. The default is
                        work, which is about puting a worker to run its
                        task.""", default="work", nargs='?', const=1)
    parser.add_argument("--worker", "-w",
                        type=str, help="The worker to execute",
                        choices=['dcm', 'dsp'])
    parser.add_argument("--dsp", "-d", type=str, default='',
                        help="The specific DSP worker to execute",
                        action=ChangeWorker)
    parser.add_argument("--generate-report", "-g", help="Generate report",
                        required=False, default=False, action='store_true')
    parser.add_argument("--port", "-p", type=int,
                        help="The port for serve the web app",
                        default=8080, required=False)
    parser.add_argument("--delay", "-t", type=int,
                        help="Delay for operation",
                        default=5, required=False)
    return parser


def manager(args):
    """this function will execute the proper commands according to the
    arguments

    Params
    ------
    args : array_like
        a list of arguments

    """
    if args.action == "init":
        logger.info("Initializing the environment")
        initialize_database()
    elif args.action == "work":
        if args.generate_report:
            logger.info("Generating report")
            generate_report()
        else:
            workers = []
            if args.worker == 'dcm':
                logger.info("Triggering DCM worker")
                workers.append(DcmWorker())
            elif args.dsp:
                logger.info("Triggering DSP worker [{}]".format(args.dsp))
                workers.append(DspWorker(args.dsp))
            else:
                logger.info("Triggering DSP workers")
                dsp_opts = BucketHelper.dsp_available()
                logger.info("Found [{}]".format(", ".join(dsp_opts)))
                for opt in dsp_opts:
                    logger.info("Creating structure for [{}]".format(opt))
                    workers.append(DspWorker(opt))
            for w in workers:
                w.extract().transform().load()

    elif args.action == "serve":
        # keep these together for sanity reasons
        try:
            logger.info("Server requested")
            from webapp.run import app
            app.run(port=args.port)
        except (KeyboardInterrupt, SystemExit):
            logger.info("Server shutdown")
        except Exception as err:
            logger.exception(err)

    elif args.action == "operate":
        logger.info("Operation requested")
        delay = args.delay
        while True:
            try:
                time.sleep(delay)
            except KeyboardInterrupt:
                break
            except Exception as err:
                logger.exception(err)
                time.sleep(2*delay)
        logger.info("Operation stopped")

    logger.info("Finish")


def main():
    """
    entry point for setup tools
    """
    parser = create_parser()
    if not sys.stdin.isatty():
        input_string = sys.stdin.read()
        args = parser.parse_args([input_string] + sys.argv[1:])
    else:
        args = parser.parse_args()
    manager(args)


if __name__ == '__main__':
    main()



# third-party imports
from flask_login import UserMixin
from werkzeug.security import generate_password_hash, check_password_hash
from sqlalchemy.sql import func
from sqlalchemy import Index

# local imports
from webapp.app import db, login_manager


class DCM(db.Model):
    """
    Create a DCM classified table
    """

    __tablename__ = 'dcm_classified'
    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.DateTime, nullable=False)
    campaign_id = db.Column(db.Integer, nullable=False)
    campaign = db.Column(db.String(75), nullable=False)
    placement_id = db.Column(db.Integer, nullable=False)
    placement = db.Column(db.String(75), nullable=False)
    impressions = db.Column(db.Float, nullable=False)
    clicks = db.Column(db.Integer, nullable=False)
    reach = db.Column(db.Float, nullable=False)
    brand = db.Column(db.String(25), nullable=False)
    sub_brand = db.Column(db.String(25), nullable=False)
    dsp = db.Column(db.String(25), nullable=False)
    created_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now())
    updated_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now(), onupdate=func.now())


# Create an index to not allow reapeated values on these dimensions
Index('dcm_classified_index', DCM.date, DCM.brand, DCM.sub_brand,
      DCM.campaign_id, DCM.campaign, DCM.placement_id, DCM.placement, DCM.dsp,
      unique=True)


class DCMRaw(db.Model):
    """
    Create a DCM raw table
    """

    __tablename__ = 'dcm_raw'
    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.DateTime, nullable=False)
    campaign_id = db.Column(db.Integer, nullable=False)
    campaign = db.Column(db.String(75), nullable=False)
    placement_id = db.Column(db.Integer, nullable=False)
    placement = db.Column(db.String(75), nullable=False)
    impressions = db.Column(db.Float, nullable=False)
    clicks = db.Column(db.Integer, nullable=False)
    reach = db.Column(db.Float, nullable=False)
    created_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now())
    updated_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now(), onupdate=func.now())


# Create an index to not allow reapeated values on these dimensions
Index('dcm_raw_index', DCMRaw.date, DCMRaw.campaign_id, DCMRaw.campaign,
      DCMRaw.placement_id, DCMRaw.placement, unique=True)


class DSP(db.Model):
    """
    Create a DSP classified table
    """

    __tablename__ = 'dsp_classified'
    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.DateTime, nullable=False)
    campaign_id = db.Column(db.Integer, nullable=False)
    campaign = db.Column(db.String(75), nullable=False)
    impressions = db.Column(db.Float, nullable=False)
    clicks = db.Column(db.Integer, nullable=False)
    cost = db.Column(db.Float, nullable=False)
    brand = db.Column(db.String(25), nullable=False)
    sub_brand = db.Column(db.String(25), nullable=False)
    dsp = db.Column(db.String(25), nullable=False)
    created_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now())
    updated_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now(), onupdate=func.now())


# Create an index to not allow reapeated values on these dimensions
Index('dsp_classified_index', DSP.date, DSP.brand, DSP.sub_brand,
      DSP.campaign_id, DSP.campaign, DSP.dsp, unique=True)


class DSPRaw(db.Model):
    """
    Create a DSP raw table
    """

    __tablename__ = 'dsp_raw'
    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.DateTime, nullable=False)
    campaign_id = db.Column(db.Integer, nullable=False)
    campaign = db.Column(db.String(75), nullable=False)
    impressions = db.Column(db.Float, nullable=False)
    clicks = db.Column(db.Integer, nullable=False)
    cost = db.Column(db.Float, nullable=False)
    created_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now())
    updated_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now(), onupdate=func.now())


# Create an index to not allow reapeated values on these dimensions
Index('dsp_raw_index', DSPRaw.date, DSPRaw.campaign_id, DSPRaw.campaign,
      unique=True)


class Report(db.Model):
    """
    Create a Report table
    """

    __tablename__ = 'report'
    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.DateTime, nullable=False)
    brand = db.Column(db.String(25), nullable=False)
    sub_brand = db.Column(db.String(25), nullable=False)
    ad_campaign_id = db.Column(db.Integer, nullable=False)
    ad_campaign = db.Column(db.String(75), nullable=False)
    dsp = db.Column(db.String(25), nullable=False)
    dsp_campaign_id = db.Column(db.Integer, nullable=False)
    dsp_campaign = db.Column(db.String(75), nullable=False)
    ad_impressions = db.Column(db.Float, nullable=False)
    ad_clicks = db.Column(db.Integer, nullable=False)
    ad_reach = db.Column(db.Float, nullable=False)
    dsp_impressions = db.Column(db.Float, nullable=False)
    dsp_clicks = db.Column(db.Integer, nullable=False)
    dsp_cost = db.Column(db.Float, nullable=False)
    created_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now())
    updated_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now(), onupdate=func.now())

    @property
    def serialize(self):
        """Return object data in serializeable format"""
        return {
            "date": self.date.isoformat(),
            "brand": self.brand,
            "sub_brand": self.sub_brand,
            "ad_campaign_id": self.ad_campaign_id,
            "ad_campaign": self.ad_campaign,
            "dsp": self.dsp,
            "dsp_campaign_id": self.dsp_campaign_id,
            "dsp_campaign": self.dsp_campaign,
            "ad_impressions": self.ad_impressions,
            "ad_clicks": self.ad_clicks,
            "ad_reach": self.ad_reach,
            "dsp_impressions": self.dsp_impressions,
            "dsp_clicks": self.dsp_clicks,
            "dsp_cost": self.dsp_cost,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat()
            }


# Create an index to not allow reapeated values on these dimensions
Index('report_index', Report.date, Report.brand, Report.sub_brand,
      Report.ad_campaign_id, Report.ad_campaign, Report.dsp,
      Report.dsp_campaign_id, Report.dsp_campaign, unique=True)


class Classification(db.Model):
    """
    Create a classification model
    """

    __tablename__ = 'classifications'
    id = db.Column(db.Integer, primary_key=True)
    pattern = db.Column(db.String(256), index=True)
    brand = db.Column(db.String(25), nullable=True)
    sub_brand = db.Column(db.String(25), nullable=True)
    dsp = db.Column(db.String(25), nullable=True)
    use_campaign_id = db.Column(db.Boolean, nullable=False, default=False)
    use_campaign = db.Column(db.Boolean, nullable=False, default=False)
    use_placement_id = db.Column(db.Boolean, nullable=False, default=False)
    use_placement = db.Column(db.Boolean, nullable=False, default=False)
    created_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now())
    updated_at = db.Column(db.DateTime, nullable=False,
                           server_default=func.now(), onupdate=func.now())


# Create an index to not allow reapeated values on these dimensions
Index('classification_index', Classification.pattern, Classification.brand,
      Classification.sub_brand, Classification.dsp,
      Classification.use_campaign, Classification.use_campaign_id,
      Classification.use_placement, Classification.use_placement_id,
      unique=True)


class User(UserMixin, db.Model):
    """
    Create an User table
    """

    # Ensures table will be named in plural and not in singular
    # as is the name of the model
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(60), index=True, unique=True)
    username = db.Column(db.String(60), index=True, unique=True)
    first_name = db.Column(db.String(60), index=True)
    last_name = db.Column(db.String(60), index=True)
    password_hash = db.Column(db.String(128))
    is_admin = db.Column(db.Boolean, default=False)

    @property
    def password(self):
        """
        Prevent pasword from being accessed
        """
        raise AttributeError('password is not a readable attribute.')

    @password.setter
    def password(self, password):
        """
        Set password to a hashed password
        """
        self.password_hash = generate_password_hash(password)

    def verify_password(self, password):
        """
        Check if hashed password matches actual password
        """
        return check_password_hash(self.password_hash, password)

    def __repr__(self):
        return '<User: {}>'.format(self.username)


# Set up user_loader
@login_manager.user_loader
def load_user(user_id):
    return User.query.get(int(user_id))
